//
//  NanoVM, a tiny java VM for the Atmel AVR family
//  Copyright (C) 2005 by Till Harbaum <Till@Harbaum.org>
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
// 

//
//  vm.c
//

#include "types.h"
#include "debug.h"
#include "config.h"
#include "error.h"

#include "vm.h"
#include "opcodes.h"
#include "native_impl.h"
#include "native.h"
#include "heap.h"
#include "nvmfile.h"
#include "stack.h"

#ifdef ARRAY
#include "array.h"
#endif

void vm_init(void) {
  DEBUGF("vm_init() with %d static fields\n", nvmfile_get_static_fields());

  // init heap
  heap_init();

  // get stack space from heap and setup stack
  stack_init(nvmfile_get_static_fields());
 
  stack_push(0); // args parameter to main (should be a string array)
}

void *vm_get_addr(u16_t ref) {
  if(!(ref & IMMEDIATE_MASK)) 
    error(ERROR_VM_ILLEGAL_REFERENCE);

  if((ref & TYPE_MASK) == TYPE_HEAP)
    return heap_get_addr(ref & ~TYPE_MASK);

  // return nvmfile address and set marker indicating
  // that this is inside the nvm file (and may have
  // to be accessed in a special manner)
  return NVMFILE_SET(nvmfile_get_addr(ref & ~TYPE_MASK));
}

// expand 15 bit immediate to 16 bits
vm_t vm_expand(vm_t val) {
  if(val & (IMMEDIATE_MASK>>1)) val |= IMMEDIATE_MASK;   // expand sign bit
  return val;
}

vm_t *locals;

// pc/methodref/localsoffset
#define VM_METHOD_CALL_REQUIREMENTS 3

// create an instance of a class. check if it's local (within 
// the nvm file) or native (implemented by the runtime environment)
void vm_new(u16_t mref) {

  if(NATIVE_ID2CLASS(mref) < NATIVE_CLASS_BASE) {
    DEBUGF("local new #%d\n", NATIVE_ID2CLASS(mref));

    DEBUGF("non static fields: %d\n", 
	       nvmfile_get_class_fields(NATIVE_ID2CLASS(mref)));

    // create object with 
    heap_id_t h = heap_alloc(TRUE, sizeof(vm_t) * 
     (VM_CLASS_CONST_ALLOC+nvmfile_get_class_fields(NATIVE_ID2CLASS(mref))));

    stack_push(TYPE_HEAP | h);

    // store reference in object, so we can later determine which kind
    // of object this is. this is required for inheritance
    ((vm_t*)heap_get_addr(h))[0] = mref;

    return;
  }

  native_new(mref);
}

// we prefetch arguments from the program storage
// and this is the type it is stored into
typedef union {
  s16_t w;
  struct {
    s08_t bl, bh;
  };
} vm_arg_t;

void   vm_run(u16_t mref) {
  u08_t instr, pc_inc, *pc;
  vm_t tmp1, tmp2;
  vm_arg_t arg0;
  nvm_method_hdr_t mhdr, *mhdr_ptr;

#ifdef STACK_CHECK
  stack_save_sp();
#endif

  DEBUGF("Running method %d\n", mref);

  // load method header into ram
  mhdr_ptr = nvmfile_get_method_hdr(mref);
  nvmfile_read(&mhdr, mhdr_ptr, sizeof(nvm_method_hdr_t));

  // determine method description address and code
  pc = (u08_t*)mhdr_ptr + mhdr.code_index;

  // make space for locals on the stack
  DEBUGF("Allocating space for %d local(s) and %d "
	     "stack elements - %d args\n", 
	     mhdr.max_locals, mhdr.max_stack, mhdr.args);
  
  // increase stack space. locals will be put on the stack as 
  // well. method arguments are part of the locals and are 
  // already on the stack
  heap_steal(sizeof(vm_t) * (mhdr.max_locals + mhdr.max_stack + mhdr.args));

  // determine address of current locals (stack pointer + 1)
  locals = stack_get_sp() + 1;
  stack_add_sp(mhdr.max_locals);
  stack_save_base();
  
  do {
    instr = nvmfile_read08(pc);
    pc_inc = 1;
    
    DEBUGF("%d/(sp:%d) - "DBG8" (%d): ", 
	       (pc-(u08_t*)mhdr_ptr) - mhdr.code_index, 
	       stack_get_depth(), instr, instr);
    
    // prefetch next args (in big endian order)
    arg0.bh = nvmfile_read08(pc+1);
    arg0.bl = nvmfile_read08(pc+2);

    if(instr == OP_NOP) {
      DEBUGF("nop\n");
    }
    
    else if(instr == OP_BIPUSH) {
      stack_push(arg0.bh); pc_inc = 2;
      DEBUGF("bipush #%d\n", stack_peek(0));
    } 

    else if(instr == OP_SIPUSH) {
      stack_push(~IMMEDIATE_MASK & (arg0.w)); pc_inc = 3;
      DEBUGF("sipush #"DBG16"\n", stack_peek_expand(0));
    } 
    
    else if((instr >= OP_ICONST_M1) && (instr <= OP_ICONST_5)) {
      stack_push(instr - OP_ICONST_0);
      DEBUGF("iconst_%d\n", stack_peek(0));
    } 

    // move integer from stack into locals
    else if(instr == OP_ISTORE) {
      locals[arg0.bh] = stack_pop(); pc_inc = 2;
      DEBUGF("istore %d (%d)\n", arg0.bh, vm_expand(locals[arg0.bh]));
    } 
    
    // move integer from stack into locals
    else if((instr >= OP_ISTORE_0) && (instr <= OP_ISTORE_3)) {
      locals[instr - OP_ISTORE_0] = stack_pop();
      DEBUGF("istore_%d (%d)\n", instr - OP_ISTORE_0, 
		 vm_expand(locals[instr - OP_ISTORE_0]));
    } 

    // load int from local variable (push local var)
    else if(instr == OP_ILOAD) {
      stack_push(locals[arg0.bh]); pc_inc = 2;
      DEBUGF("iload %d (%d, "DBG16")\n", locals[arg0.bh], 
		 stack_peek_expand(0), stack_peek_expand(0));
    } 

    // push local onto stack
    else if((instr >= OP_ILOAD_0) && (instr <= OP_ILOAD_3)) {
      stack_push(locals[instr - OP_ILOAD_0]);
      DEBUGF("iload_%d (%d, "DBG16")\n", instr-OP_ILOAD_0, 
		 stack_peek_expand(0), stack_peek_expand(0));
    } 

    // immediate comparison / comparison with zero
    else if((instr >= OP_IFEQ) && (instr <= OP_IF_ICMPLE)) {
      DEBUGF("if");

      if((instr >= OP_IFEQ) && (instr <= OP_IFLE)) {
	// comparision with zero
	tmp2 = 0;
	instr -= OP_IFEQ - OP_IF_ICMPEQ;
      } else {
	// comparison with second argument
	DEBUGF("_cmp");
	tmp2 = stack_pop_expand();
      }

      tmp1 = stack_pop_expand();

      switch(instr) {
	case OP_IF_ICMPEQ: DEBUGF("eq %d", arg0.w); 
	  tmp1 = (tmp1 == tmp2); break;
	case OP_IF_ICMPNE: DEBUGF("ne %d", arg0.w); 
	  tmp1 = (tmp1 != tmp2); break;
	case OP_IF_ICMPLT: DEBUGF("lt %d", arg0.w); 
	  tmp1 = (tmp1 <  tmp2); break;
	case OP_IF_ICMPGE: DEBUGF("ge %d", arg0.w); 
	  tmp1 = (tmp1 >= tmp2); break;
	case OP_IF_ICMPGT: DEBUGF("gt %d", arg0.w); 
	  tmp1 = (tmp1 >  tmp2); break;
	case OP_IF_ICMPLE: DEBUGF("le %d", arg0.w); 
	  tmp1 = (tmp1 <= tmp2); break;	    
      }
      
      // change pc if jump has been taken
      if(tmp1) { DEBUGF(" -> taken\n"); pc += arg0.w; pc_inc = 0; }
      else     { DEBUGF(" -> not taken\n"); pc_inc = 3; }
    } 

    else if(instr == OP_GOTO) {
      pc_inc = 3;
      DEBUGF("goto %d\n", arg0.w); 
      pc += (arg0.w-3);
    } 

    // two operand arithmetic
    else if((instr >= OP_IADD) && (instr <= OP_IINC)) {
      // single operand arithmetic
      if(instr == OP_INEG) {
	arg0.w = -stack_pop_expand();
	DEBUGF("neg(%d)\n", vm_expand(~IMMEDIATE_MASK & -arg0.w));
	stack_push(~IMMEDIATE_MASK & arg0.w);

      } else if(instr == OP_IINC) {
	DEBUGF("iinc %d,%d\n", arg0.bh, arg0.bl); 
	locals[arg0.bh] = (vm_expand(locals[arg0.bh]) + arg0.bl) 
	  & ~IMMEDIATE_MASK; 
	pc_inc = 3;

      } else {
	tmp1 = stack_pop_expand();  // fetch operands from stack
	arg0.w = stack_pop_expand();
	
	switch(instr) {
	  case OP_IADD:  DEBUGF("iadd(%d,%d)", arg0.w, tmp1); 
	    arg0.w  += tmp1; break;
	  case OP_ISUB:  DEBUGF("isub(%d,%d)", arg0.w, tmp1); 
	    arg0.w  -= tmp1; break;
	  case OP_IMUL:  DEBUGF("imul(%d,%d)", arg0.w, tmp1); 
	    arg0.w  *= tmp1; break;
	  case OP_IDIV:  DEBUGF("idiv(%d,%d)", arg0.w, tmp1); 
	    if(!tmp1) error(ERROR_VM_DIVISION_BY_ZERO);
	    arg0.w  /= tmp1; break;
	  case OP_IREM:  DEBUGF("irem(%d,%d)", arg0.w, tmp1); 
	    arg0.w  %= tmp1; break;
	  case OP_ISHL:  DEBUGF("ishl(%d,%d)", arg0.w, tmp1); 
	    arg0.w <<= tmp1; break;
	  case OP_ISHR:  DEBUGF("ishr(%d,%d)", arg0.w, tmp1); 
	    arg0.w >>= tmp1; break;
	  case OP_IAND:  DEBUGF("iand(%d,%d)", arg0.w, tmp1); 
	    arg0.w  &= tmp1; break;
	  case OP_IOR:   DEBUGF("ior(%d,%d)",  arg0.w, tmp1); 
	    arg0.w  |= tmp1; break;
	  case OP_IXOR:  DEBUGF("ixor(%d,%d)", arg0.w, tmp1); 
	    arg0.w  ^= tmp1; break;
	  case OP_IUSHR: DEBUGF("iushr(%d,%d)",arg0.w, tmp1); 
	    arg0.w = ((vmu_t)arg0.w >> tmp1); break;
	}
	
	// and finally push result
	DEBUGF(" = %d\n", vm_expand(~IMMEDIATE_MASK & arg0.w));
	stack_push(~IMMEDIATE_MASK & arg0.w);
      }
    }

    else if((instr == OP_IRETURN) || (instr == OP_RETURN)) {
      if(instr == OP_IRETURN) {
	arg0.w = stack_pop();     // save result
	DEBUGF("i");
      }
      
      DEBUGF("return: ");

      // return from locally called method? other case: return
      // from main() -> end of program
      if(!stack_is_empty()) {
	u08_t old_locals = mhdr.max_locals;
	u08_t old_unsteal = VM_METHOD_CALL_REQUIREMENTS +
	  mhdr.max_locals + mhdr.max_stack + mhdr.args;
	u16_t old_localsoffset = stack_pop();
	
	// make space for locals on the stack
	DEBUGF("Return from method with %d local(s) and %d "
		   "stack elements - %d args\n", 
		   mhdr.max_locals, mhdr.max_stack, mhdr.args);
	
	mref = stack_pop();
	
	// read header of method to return to
	mhdr_ptr = nvmfile_get_method_hdr(mref);
	// load method header into ram
	nvmfile_read(&mhdr, mhdr_ptr, sizeof(nvm_method_hdr_t));
	
	// restore pc
	pc = (void*)mhdr_ptr + stack_pop();
	pc_inc = 3; // continue _behind_ calling invoke instruction
	
	// and remove locals from stack and hope that method left
	// an uncorrupted stack
	stack_add_sp(-old_locals);
	locals = stack_get_sp() - old_localsoffset;
	
	// give memory used by returning method back to heap
	heap_unsteal(sizeof(vm_t) * old_unsteal);
	
	if(instr == OP_IRETURN) {
	  DEBUGF("ireturn val: %d\n", arg0.w);
	  stack_push(arg0.w);
	}

	instr = OP_NOP;  // make vm continue
      }
    }

    // discard both top stack items
    else if(instr == OP_POP2) {
      DEBUGF("ipop\n");
      stack_pop(); stack_pop();
    }
    
    // discard top stack item
    else if(instr == OP_POP) {
      DEBUGF("pop\n");
      stack_pop();
    }
    
    // duplicate top stack item
    else if(instr == OP_DUP) {
      stack_push(stack_peek(0));
      DEBUGF("dup ("DBG16")\n", stack_peek(0) & 0xffff);
    }

    // duplicate top two stack items  (a,b -> a,b,a,b)
    else if(instr == OP_DUP2) {
      stack_push(stack_peek(1));
      stack_push(stack_peek(1));
      DEBUGF("dup2 ("DBG16","DBG16")\n", 
	     stack_peek(0) & 0xffff, stack_peek(1) & 0xffff);
    }
    
#ifdef SWITCH
    else if(instr == OP_TABLESWITCH) {
      u08_t skip = arg0.bh;

      DEBUGF("skip: %d\n", skip);
      tmp1 = ((nvmfile_read08(pc+skip+7)<<8) | 
	      nvmfile_read08(pc+skip+8));        // get low value
      tmp2 = ((nvmfile_read08(pc+skip+11)<<8) | 
	      nvmfile_read08(pc+skip+12));       // get high value
      arg0.w = stack_pop();                              // get actual value
      DEBUGF("tableswitch %d-%d (%d)\n", tmp1, tmp2, arg0.w);
      
      // value within range?
      if((arg0.w < tmp1)||(arg0.w > tmp2))
	// no: use default
	tmp2 = skip + 3; 	  
      else
	// yes: get offset from table
	tmp2 = skip + 3 + 12 + ((arg0.w - tmp1)<<2);	
      
      // and do the jump
      pc += ((nvmfile_read08(pc+tmp2+0)<<8) | 
	     nvmfile_read08(pc+tmp2+1));
      pc_inc = 0;
    }
#endif
    
    // get static field from class
    else if(instr == OP_GETSTATIC) {
      pc_inc = 3;   // prefetched data used
      DEBUGF("getstatic #"DBG16"\n", arg0.w);
      stack_push(stack_get_static(arg0.w));
    }
    
    else if(instr == OP_PUTSTATIC) {
      pc_inc = 3;
      stack_set_static(arg0.w, stack_pop());
      DEBUGF("putstatic #"DBG16" -> "DBG16"\n", 
	     arg0.w, stack_get_static(arg0.w));
    }
    
    // push item from constant pool
    else if(instr == OP_LDC) {
      pc_inc = 2;
      DEBUGF("ldc #"DBG16"\n", arg0.bh);
      stack_push(TYPE_CONST | arg0.bh);
    }
    
    else if((instr >= OP_INVOKEVIRTUAL)&&(instr <= OP_INVOKESTATIC)) {
      DEBUGF("invoke");

#ifdef DEBUG
      if(instr == OP_INVOKEVIRTUAL) { DEBUGF("virtual"); }
      if(instr == OP_INVOKESPECIAL) { DEBUGF("special"); }
      if(instr == OP_INVOKESTATIC)  { DEBUGF("static"); }
#endif

      DEBUGF(" #"DBG16"\n", 0xffff & arg0.w);
      
      // invoke a method. check if it's local (within the nvm file)
      // or native (implemented by the runtime environment)
      if(arg0.bh < NATIVE_CLASS_BASE) {
	DEBUGF("local method call from method %d to %d\n", mref, arg0.w);

	// save current pc (relative to method start)
	tmp1 = (void*)pc-(void*)mhdr_ptr;
	
	// get pointer to new method
	mhdr_ptr = nvmfile_get_method_hdr(arg0.w);
	
	// load new method header into ram
	nvmfile_read(&mhdr, mhdr_ptr, sizeof(nvm_method_hdr_t));
	
#ifdef INHERITANCE
	// check class on stack. it may be not the one we expect.
	// this happens due to inheritance
	if(instr == OP_INVOKEVIRTUAL) { 
	  DEBUGF("checking inheritance\n");

	  // fetch class reference from stack and use it to address
	  // the class instance on the heap. The first entry in this 
	  // object is the class id of it
	  vm_t mref = ((vm_t*)heap_get_addr(stack_peek(0) & ~TYPE_MASK))[0];
	  DEBUGF("class ref on stack/ref: %d/%d\n", 
		     NATIVE_ID2CLASS(mref), NATIVE_ID2CLASS(mhdr.id));

	  if(NATIVE_ID2CLASS(mref) != NATIVE_ID2CLASS(mhdr.id)) {
	    DEBUGF("stack/ref class mismatch -> inheritance\n");

	    // get matching method in class on stack or its
	    // super classes
	    arg0.bl = nvmfile_get_method_by_class_and_id(
	      NATIVE_ID2CLASS(mref), NATIVE_ID2METHOD(mhdr.id));

	    // get pointer to new method
	    mhdr_ptr = nvmfile_get_method_hdr(arg0.bl);
	
	    // load new method header into ram
	    nvmfile_read(&mhdr, mhdr_ptr, sizeof(nvm_method_hdr_t));
	  }
	}
#endif
	
	// arguments are left on the stack by the calling
	// method and expected in the locals by the called
	// method. Thus we make this part of the old stack
	// be the locals part of the method
	DEBUGF("Remove %d args from stack\n", mhdr.args);
	stack_add_sp(-mhdr.args);
	
	tmp2 = stack_get_sp() - locals;
	
	locals = stack_get_sp() + 1;
	
#ifdef DEBUG
	if(instr == OP_INVOKEVIRTUAL) { 
	  DEBUGF("virtual call with object reference "DBG16"\n",
		     locals[0]); 
	}
#endif

	// make space for locals on the stack
	DEBUGF("Allocating space for %d local(s) and %d "
		   "stack elements - %d args\n", 
		   mhdr.max_locals, mhdr.max_stack, mhdr.args);
	
	// increase stack space. locals will be put on the stack as 
	// well. method arguments are part of the locals and are 
	// already on the stack
	heap_steal(sizeof(vm_t) *
		   (VM_METHOD_CALL_REQUIREMENTS +
		    mhdr.max_locals + mhdr.max_stack + mhdr.args));
	
	// add space for locals on stack
	stack_add_sp(mhdr.max_locals);
	
	// push everything required to return onto the stack
	stack_push(tmp1);   // pc offset
	stack_push(mref);   // method reference
	stack_push(tmp2);   // locals offset
	
	// set new pc (this is the actual call)
	mref = arg0.w;
	pc = (u08_t*)mhdr_ptr + mhdr.code_index;
	pc_inc = 0;  // don't add further bytes to program counter
      } else { 
	native_invoke(arg0.w);
	pc_inc = 3;   // prefetched data used
      }
    }
    
    else if(instr == OP_GETFIELD) {
      pc_inc = 3;
      DEBUGF("getfield #%d\n", arg0.w);
      stack_push(((vm_t*)heap_get_addr(stack_pop() & ~TYPE_MASK))
	      [VM_CLASS_CONST_ALLOC+arg0.w]);
    }
    
    else if(instr == OP_PUTFIELD) {
      pc_inc = 3;
      tmp1 = stack_pop();
      
      DEBUGF("putfield #%d\n", arg0.w);
      ((vm_t*)heap_get_addr(stack_pop() & ~TYPE_MASK))
	[VM_CLASS_CONST_ALLOC+arg0.w] = tmp1;
    }
    
    else if(instr == OP_NEW) {
      pc_inc = 3;
      DEBUGF("new #"DBG16"\n", 0xffff & arg0.w);
      vm_new(arg0.w);
    }
    
#ifdef ARRAY
    else if(instr == OP_NEWARRAY) {
      pc_inc = 2;
      stack_push(array_new(stack_pop(), arg0.bh) | TYPE_HEAP);
    }
    
    else if(instr == OP_ARRAYLENGTH) {
      stack_push(array_length(stack_pop() & ~TYPE_MASK));
    }
    
    else if(instr == OP_BASTORE) {
      arg0.w = stack_pop_expand();       // value
      tmp1 = stack_pop_expand();         // index
      // third parm on stack: array reference
      array_bastore(stack_pop() & ~TYPE_MASK, tmp1, arg0.w);
    }
    
    else if(instr == OP_IASTORE) {
      arg0.w = stack_pop_expand();       // value
      tmp1 = stack_pop_expand();         // index
      // third parm on stack: array reference
      array_iastore(stack_pop() & ~TYPE_MASK, tmp1, arg0.w);
    }
    
    else if(instr == OP_BALOAD) {
      arg0.w = stack_pop_expand();       // index
      // second parm on stack: array reference
      stack_push(array_baload(stack_pop() & ~TYPE_MASK, arg0.w));
    }
    
    else if(instr == OP_IALOAD) {
      arg0.w = stack_pop_expand();       // index
      // second parm on stack: array reference
      stack_push(array_iaload(stack_pop() & ~TYPE_MASK, arg0.w));
    }
#endif
    
    else {
      error(ERROR_VM_UNSUPPORTED_OPCODE);
    }
    
    // reset watchdog here if present

    pc += pc_inc;
  } while((instr != OP_IRETURN)&&(instr != OP_RETURN));

  // and remove locals from stack and hope that method left
  // an uncorrupted stack
  stack_add_sp(-mhdr.max_locals);

#ifdef STACK_CHECK
  stack_verify_sp();
#endif

  // give memory back to heap
  heap_unsteal(sizeof(vm_t) * (mhdr.max_locals + mhdr.max_stack + mhdr.args));
}

